@@ -26,6 +26,7 @@
 #define NDPI_CURRENT_PROTO NDPI_PROTOCOL_DNS
 
 #include "ndpi_api.h"
+#include "ndpi_private.h"
 
 #define FLAGS_MASK 0x8000
 
@@ -37,16 +38,13 @@
 
 #define PKT_LEN_ALERT 512
 
-/* ndpi_main.c */
-extern u_int8_t ndpi_iph_is_valid_and_not_fragmented(const struct ndpi_iphdr *iph, const u_int16_t ipsize);
 
 static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct,
 			    struct ndpi_flow_struct *flow);
 
 /* *********************************************** */
 
-static void ndpi_check_dns_type(struct ndpi_detection_module_struct *ndpi_struct,
-				struct ndpi_flow_struct *flow,
+static void ndpi_check_dns_type(struct ndpi_flow_struct *flow,
 				u_int16_t dns_type) {
   /* https://en.wikipedia.org/wiki/List_of_DNS_record_types */
 
@@ -94,7 +92,7 @@ static void ndpi_check_dns_type(struct ndpi_detection_module_struct *ndpi_struct
   case 106:
   case 107:
   case 259:
-    ndpi_set_risk(ndpi_struct, flow, NDPI_DNS_SUSPICIOUS_TRAFFIC, "Obsolete DNS record type");
+    ndpi_set_risk(flow, NDPI_DNS_SUSPICIOUS_TRAFFIC, "Obsolete DNS record type");
     break;
   }
 }
@@ -254,7 +252,7 @@ static u_int8_t ndpi_grab_dns_name(struct ndpi_packet_struct *packet,
 	  
 	  hostname_is_valid = 0;
 	  
-	  if (isprint(c) == 0) {
+	  if (ndpi_isprint(c) == 0) {
 	    _hostname[j++] = '?';
 	  } else {
 	    _hostname[j++] = '_';
@@ -322,7 +320,8 @@ static int search_valid_dns(struct ndpi_detection_module_struct *ndpi_struct,
 	  x++;
       }
     } else {
-      ndpi_set_risk(ndpi_struct, flow, NDPI_MALFORMED_PACKET, "Invalid DNS Header");
+      if(flow->detected_protocol_stack[0] != NDPI_PROTOCOL_UNKNOWN)
+        ndpi_set_risk(flow, NDPI_MALFORMED_PACKET, "Invalid DNS Header");
       return(1 /* invalid */);
     }
   } else {
@@ -352,10 +351,10 @@ static int search_valid_dns(struct ndpi_detection_module_struct *ndpi_struct,
 
       snprintf(str, sizeof(str), "DNS Error Code %s",
 	       dns_error_code2string(flow->protos.dns.reply_code, buf, sizeof(buf)));
-      ndpi_set_risk(ndpi_struct, flow, NDPI_ERROR_CODE_DETECTED, str);
+      ndpi_set_risk(flow, NDPI_ERROR_CODE_DETECTED, str);
     } else {
-      if(ndpi_isset_risk(ndpi_struct, flow, NDPI_SUSPICIOUS_DGA_DOMAIN)) {
-	ndpi_set_risk(ndpi_struct, flow, NDPI_RISKY_DOMAIN, "DGA Name Query with no Error Code");
+      if(ndpi_isset_risk(flow, NDPI_SUSPICIOUS_DGA_DOMAIN)) {
+	ndpi_set_risk(flow, NDPI_RISKY_DOMAIN, "DGA Name Query with no Error Code");
       }
     }
 
@@ -432,7 +431,7 @@ static int search_valid_dns(struct ndpi_detection_module_struct *ndpi_struct,
 	  rsp_ttl  = ntohl(*((u_int32_t*)&packet->payload[x+2]));
 
 	  if(rsp_ttl == 0)
-	    ndpi_set_risk(ndpi_struct, flow, NDPI_MINOR_ISSUES, "DNS Record with zero TTL");
+	    ndpi_set_risk(flow, NDPI_MINOR_ISSUES, "DNS Record with zero TTL");
 	  
 #ifdef DNS_DEBUG
 	  printf("[DNS] TTL = %u\n", rsp_ttl);
@@ -440,12 +439,14 @@ static int search_valid_dns(struct ndpi_detection_module_struct *ndpi_struct,
 #endif
 
 	  if(found == 0) {
-	    ndpi_check_dns_type(ndpi_struct, flow, rsp_type);
+	    ndpi_check_dns_type(flow, rsp_type);
 	    flow->protos.dns.rsp_type = rsp_type;
 	  }
 	  
 	  /* x points to the response "class" field */
 	  if((x+12) <= packet->payload_packet_len) {
+	    u_int32_t ttl = ntohl(*((u_int32_t*)&packet->payload[x+2]));
+	    
 	    x += 6;
 	    data_len = get16(&x, packet->payload);
 
@@ -474,15 +475,34 @@ static int search_valid_dns(struct ndpi_detection_module_struct *ndpi_struct,
 			 || ((rsp_type == 0x1c) && (data_len == 16)) /* AAAA */
 			 )) {
 		if(found == 0) {
-		  memcpy(&flow->protos.dns.rsp_addr, packet->payload + x, data_len);
-		  found = 1;
+		  
+		  if(flow->protos.dns.num_rsp_addr < MAX_NUM_DNS_RSP_ADDRESSES) {
+		    /* Necessary for IP address comparison */
+		    memset(&flow->protos.dns.rsp_addr[flow->protos.dns.num_rsp_addr], 0, sizeof(ndpi_ip_addr_t));
+		  
+		    memcpy(&flow->protos.dns.rsp_addr[flow->protos.dns.num_rsp_addr], packet->payload + x, data_len);
+		    flow->protos.dns.is_rsp_addr_ipv6[flow->protos.dns.num_rsp_addr] = (data_len == 16) ? 1 : 0;
+		    flow->protos.dns.rsp_addr_ttl[flow->protos.dns.num_rsp_addr] = ttl;
+
+		    if(ndpi_struct->cfg.address_cache_size)
+		      ndpi_cache_address(ndpi_struct,
+				         flow->protos.dns.rsp_addr[flow->protos.dns.num_rsp_addr],
+				         flow->host_server_name,
+				         packet->current_time_ms/1000,
+				         flow->protos.dns.rsp_addr_ttl[flow->protos.dns.num_rsp_addr]);
+
+		    ++flow->protos.dns.num_rsp_addr;
+		  }
+
+		  if(flow->protos.dns.num_rsp_addr >= MAX_NUM_DNS_RSP_ADDRESSES)
+		    found = 1;
 		}
 	      }
-
+	      
 	      x += data_len;
 	    }
 	  }
-
+	  
 	  if(found && (dns_header->additional_rrs == 0)) {
 	    /*
 	      In case we have RR we need to iterate
@@ -700,8 +720,8 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
     u_int num_queries, idx;
     char _hostname[256];
 
-    ret.master_protocol = NDPI_PROTOCOL_UNKNOWN;
-    ret.app_protocol    = (d_port == LLMNR_PORT) ? NDPI_PROTOCOL_LLMNR : ((d_port == MDNS_PORT) ? NDPI_PROTOCOL_MDNS : NDPI_PROTOCOL_DNS);
+    ret.proto.master_protocol = NDPI_PROTOCOL_UNKNOWN;
+    ret.proto.app_protocol    = (d_port == LLMNR_PORT) ? NDPI_PROTOCOL_LLMNR : (((d_port == MDNS_PORT) && isLLMNRMulticastAddress(packet) ) ? NDPI_PROTOCOL_MDNS : NDPI_PROTOCOL_DNS);
 
     if(invalid) {
       NDPI_EXCLUDE_PROTO(ndpi_struct, flow);
@@ -756,7 +776,7 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
 #ifdef DNS_DEBUG
 	printf("[DNS] Invalid query len [%u >= %u]\n", i+4, packet->payload_packet_len);
 #endif
-	ndpi_set_risk(ndpi_struct, flow, NDPI_MALFORMED_PACKET, "Invalid DNS Query Lenght");
+	ndpi_set_risk(flow, NDPI_MALFORMED_PACKET, "Invalid DNS Query Lenght");
 	break;
       } else {
 	idx = i+5, num_queries++;
@@ -765,44 +785,70 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
 
     u_int8_t hostname_is_valid = ndpi_grab_dns_name(packet, &off, _hostname, sizeof(_hostname), &len, is_mdns);
 
-    ndpi_hostname_sni_set(flow, (const u_int8_t *)_hostname, len);
+    ndpi_hostname_sni_set(flow, (const u_int8_t *)_hostname, len, is_mdns ? NDPI_HOSTNAME_NORM_LC : NDPI_HOSTNAME_NORM_ALL);
 
     if (hostname_is_valid == 0)
-      ndpi_set_risk(ndpi_struct, flow, NDPI_INVALID_CHARACTERS, NULL);
+      ndpi_set_risk(flow, NDPI_INVALID_CHARACTERS, "Invalid chars detected in domain name");
 
-    dot = strchr(_hostname, '.');
-    if(dot) {
-      uintptr_t first_element_len = dot - _hostname;
+    /* Ignore reverse DNS queries */
+    if(strstr(_hostname, ".in-addr.") == NULL) {
+      dot = strchr(_hostname, '.');
+      
+      if(dot) {
+	uintptr_t first_element_len = dot - _hostname;
 
-      if(first_element_len > 32) {
-	/*
-	  The lenght of the first element in the query is very long
-	  and this might be an issue or indicate an exfiltration
-	*/
+	if((first_element_len > 48) && (!is_mdns)) {
+	  /*
+	    The lenght of the first element in the query is very long
+	    and this might be an issue or indicate an exfiltration
+	  */
 
-	/* printf("**** %lu [%s][%s]\n", first_element_len, dot, _hostname); */
-	ndpi_set_risk(ndpi_struct, flow, NDPI_DNS_SUSPICIOUS_TRAFFIC, NULL);
+	  if(ends_with(ndpi_struct, _hostname, "multi.surbl.org")
+	     || ends_with(ndpi_struct, _hostname, "spamhaus.org")
+	     || ends_with(ndpi_struct, _hostname, "rackcdn.com")
+	     || ends_with(ndpi_struct, _hostname, "akamaiedge.net")
+	     || ends_with(ndpi_struct, _hostname, "mx-verification.google.com")
+	     || ends_with(ndpi_struct, _hostname, "amazonaws.com")
+	     )
+	    ; /* Check common domain exceptions [TODO: if the list grows too much use a different datastructure] */
+	  else
+	    ndpi_set_risk(flow, NDPI_DNS_SUSPICIOUS_TRAFFIC, "Long DNS host name");
+	}
       }
     }
     
     if(len > 0) {
-      ndpi_protocol_match_result ret_match;
+      if(ndpi_struct->cfg.dns_subclassification_enabled) {
+        ndpi_protocol_match_result ret_match;
 
-      ret.app_protocol = ndpi_match_host_subprotocol(ndpi_struct, flow,
-						     flow->host_server_name,
-						     strlen(flow->host_server_name),
-						     &ret_match,
-						     NDPI_PROTOCOL_DNS);
+        ret.proto.app_protocol = ndpi_match_host_subprotocol(ndpi_struct, flow,
+						       flow->host_server_name,
+						       strlen(flow->host_server_name),
+						       &ret_match,
+						       NDPI_PROTOCOL_DNS);
+        /* Add to FPC DNS cache */
+        if(ret.proto.app_protocol != NDPI_PROTOCOL_UNKNOWN &&
+           (flow->protos.dns.rsp_type == 0x1 || flow->protos.dns.rsp_type == 0x1c) && /* A, AAAA */
+           ndpi_struct->fpc_dns_cache) {
+            ndpi_lru_add_to_cache(ndpi_struct->fpc_dns_cache,
+                                  fpc_dns_cache_key_from_dns_info(flow), ret.proto.app_protocol,
+                                  ndpi_get_current_time(flow));
+        }
+
+        if(ret.proto.app_protocol == NDPI_PROTOCOL_UNKNOWN)
+	  ret.proto.master_protocol = checkDNSSubprotocol(s_port, d_port);
+        else
+	  ret.proto.master_protocol = NDPI_PROTOCOL_DNS;
+
+        ndpi_check_dga_name(ndpi_struct, flow, flow->host_server_name, 1, 0);
+      } else {
+        ret.proto.master_protocol = checkDNSSubprotocol(s_port, d_port);
+        ret.proto.app_protocol = NDPI_PROTOCOL_UNKNOWN;
+      }
 
       /* Category is always NDPI_PROTOCOL_CATEGORY_NETWORK, regardless of the subprotocol */
       flow->category = NDPI_PROTOCOL_CATEGORY_NETWORK;
 
-      if(ret.app_protocol == NDPI_PROTOCOL_UNKNOWN)
-	ret.master_protocol = checkDNSSubprotocol(s_port, d_port);
-      else
-	ret.master_protocol = NDPI_PROTOCOL_DNS;
-
-      ndpi_check_dga_name(ndpi_struct, flow, flow->host_server_name, 1, 0);
     }
 
     /* Report if this is a DNS query or reply */
@@ -810,13 +856,15 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
 
     if(is_query) {
       /* In this case we say that the protocol has been detected just to let apps carry on with their activities */
-      ndpi_set_detected_protocol(ndpi_struct, flow, ret.app_protocol, ret.master_protocol, NDPI_CONFIDENCE_DPI);
+      ndpi_set_detected_protocol(ndpi_struct, flow, ret.proto.app_protocol, ret.proto.master_protocol, NDPI_CONFIDENCE_DPI);
 
-      /* We have never triggered extra-dissection for LLMNR. Keep the old behaviour */
-      if(ret.master_protocol != NDPI_PROTOCOL_LLMNR) {
-        /* Don't use just 1 as in TCP DNS more packets could be returned (e.g. ACK). */
-        flow->max_extra_packets_to_check = 5;
-        flow->extra_packets_func = search_dns_again;
+      if(ndpi_struct->cfg.dns_parse_response_enabled) {
+        /* We have never triggered extra-dissection for LLMNR. Keep the old behaviour */
+        if(ret.proto.master_protocol != NDPI_PROTOCOL_LLMNR) {
+          /* Don't use just 1 as in TCP DNS more packets could be returned (e.g. ACK). */
+          flow->max_extra_packets_to_check = 5;
+          flow->extra_packets_func = search_dns_again;
+	}
       }
       return; /* The response will set the verdict */
     }
@@ -837,7 +885,7 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
 	 matched a subprotocol
       **/
       NDPI_LOG_INFO(ndpi_struct, "found DNS\n");
-      ndpi_set_detected_protocol(ndpi_struct, flow, ret.app_protocol, ret.master_protocol, NDPI_CONFIDENCE_DPI);
+      ndpi_set_detected_protocol(ndpi_struct, flow, ret.proto.app_protocol, ret.proto.master_protocol, NDPI_CONFIDENCE_DPI);
     } else {
       if((flow->detected_protocol_stack[0] == NDPI_PROTOCOL_DNS)
 	 || (flow->detected_protocol_stack[1] == NDPI_PROTOCOL_DNS))
@@ -860,7 +908,7 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
       char str[48];
 
       snprintf(str, sizeof(str), "%u Bytes DNS Packet", packet->payload_packet_len);
-      ndpi_set_risk(ndpi_struct, flow, NDPI_DNS_LARGE_PACKET, str);
+      ndpi_set_risk(flow, NDPI_DNS_LARGE_PACKET, str);
     }
 
     if(packet->iph != NULL) {
@@ -869,15 +917,15 @@ static void ndpi_search_dns(struct ndpi_detection_module_struct *ndpi_struct, st
 
       /* 0: fragmented; 1: not fragmented */
       if((flags & 0x20)
-	 || (ndpi_iph_is_valid_and_not_fragmented(packet->iph, packet->l3_packet_len) == 0)) {
-	ndpi_set_risk(ndpi_struct, flow, NDPI_DNS_FRAGMENTED, NULL);
+	 || (iph_is_valid_and_not_fragmented(packet->iph, packet->l3_packet_len) == 0)) {
+	ndpi_set_risk(flow, NDPI_DNS_FRAGMENTED, NULL);
       }
     } else if(packet->iphv6 != NULL) {
       /* IPv6 */
       const struct ndpi_ip6_hdrctl *ip6_hdr = &packet->iphv6->ip6_hdr;
 
       if(ip6_hdr->ip6_un1_nxt == 0x2C /* Next Header: Fragment Header for IPv6 (44) */) {
-	ndpi_set_risk(ndpi_struct, flow, NDPI_DNS_FRAGMENTED, NULL);
+	ndpi_set_risk(flow, NDPI_DNS_FRAGMENTED, NULL);
       }
     }
   }
